------------------------------------------------------------------------------
--  ------------------         M a R T E   O S         -------------------  --
------------------------------------------------------------------------------
--                                                             V1.8     080922
--
--      'K e r n e l . T i m e d _ H a n d l e r s .
--                                       A d a _ T i m i n g _ E v e n t s'
--
--                                   Body
--
--
--
--  File 'kernel-timed_handlers-ada_timing_events.adb'                 By MAR.
--
--  Support for Ada 2005 standard packages 'Ada.Execution_Time.Timers' and
--  'Ada.Timing_Events'.
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2000-2008, Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael Gonzalez Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
------------------------------------------------------------------------------
with System;
with Ada.Unchecked_Conversion;

with Interfaces.C;

with MaRTE.Kernel.Timed_Handlers.Internals;
with MaRTE.Kernel.Timed_Handlers.Operations;
with MaRTE.Configuration_Parameters;
with MaRTE.Kernel.Debug;

package body MaRTE.Kernel.Timed_Handlers.Ada_Timing_Events is

   package THO renames K.Timed_Handlers.Operations;
   package THI renames K.Timed_Handlers.Internals;
   package DBG renames K.Debug;

   use type Interfaces.C.int;

   type Base_Timing_Event_Ac is access all Base_Timing_Event;

   -------------------------------
   --  Generic_Handler_Wrapper  --
   -------------------------------
   procedure Generic_Handler_Wrapper (Area : in System.Address;
                                      TE   : in TH.Timed_Handler_TE_Ac);
   procedure Generic_Handler_Wrapper (Area : in System.Address;
                                      TE   : in TH.Timed_Handler_TE_Ac) is
      function UC is new
        Ada.Unchecked_Conversion (System.Address, Base_Timing_Event_Ac);
   begin
      if MaRTE.Configuration_Parameters.Not_Ada_User_Code'First then
         pragma Debug (DBG.Assert (False));
         return;
      end if;

      if UC (Area).Handler /= null then
         UC (Area).Handler (UC (Area).all);  --  Call user's handler
      else
         pragma Debug (DBG.Assert (UC (Area).Based_On_Group_Clock));
         --  Only group budget without handler can expire, this should never
         --  happend for timing events and Execution Time timers since they are
         --  cleared when the handler is null

         null;
      end if;

      --  Clear event (Group budgets are not cleared after expiration)

      if not UC (Area).Based_On_Group_Clock then
         UC (Area).Handler := null;
      end if;

   end Generic_Handler_Wrapper;

   ------------------
   --  Initialize  --
   ------------------
   procedure Initialize (Event : in out Base_Timing_Event) is
      Ret : Interfaces.C.int;
   begin
      if MaRTE.Configuration_Parameters.Not_Ada_User_Code'First then
         pragma Debug (DBG.Assert (False));
         return;
      end if;

      Event.Handler := null;
      Ret :=
        THO.MaRTE_Timed_Handler_Init (Event.Timed_Handler'Unrestricted_Access,
                                      Event.Base_Clock,
                                      Generic_Handler_Wrapper'Access,
                                      Event'Address,
                                      System.Address'Size);
      pragma Debug (DBG.Assert (Ret = 0));
   end Initialize;

   ----------------
   --  Finalize  --
   ----------------
   procedure Finalize (Event : in out Base_Timing_Event) is
      Ret : Interfaces.C.int;
   begin
      if MaRTE.Configuration_Parameters.Not_Ada_User_Code'First then
         pragma Debug (DBG.Assert (False));
         return;
      end if;
      pragma Debug (DBG.Assert (not Event.Based_On_Group_Clock));

      Ret := THO.MaRTE_Timed_Handler_Destroy
        (Event.Timed_Handler'Unrestricted_Access);
      pragma Debug (DBG.Assert (Ret = 0));
      Event.Handler := null;
   end Finalize;

   --------------------------
   --  Set_Handler_HWTime  --
   --------------------------
   procedure Set_Handler_HWTime (Event   : in out Base_Timing_Event;
                                 T       : in HAL.HWTime;
                                 Handler : in Base_Timing_Event_Handler;
                                 Options : in TH.Timed_Handler_Options) is
      Ret : Interfaces.C.int;
      Flags : Integer;
   begin
      if MaRTE.Configuration_Parameters.Not_Ada_User_Code'First then
         pragma Debug (DBG.Assert (False));
         return;
      end if;
      --  Disable event
      Ret := THO.MaRTE_Timed_Handler_Disable
        (Event.Timed_Handler'Unrestricted_Access);
      pragma Debug (DBG.Assert (Ret = 0));

      --  Set new handler

      Event.Handler := Handler;

      if Handler = null and not Event.Based_On_Group_Clock then
         --  Event remains disabled (not for events based on group clocks since
         --  they are set even if its handler is null)
         return;
      end if;

      Event.Timed_Handler.T := T;
      Event.Timed_Handler.Options := Options;

      K.Enter_Critic_Section (Flags);
      THO.Set (Event.Timed_Handler, HAL.Get_HWTime);
      K.Leave_Critic_Section (Flags);

      --  re-enable event
      Ret := THO.MaRTE_Timed_Handler_Enable
        (Event.Timed_Handler'Unrestricted_Access);
      pragma Debug (DBG.Assert (Ret = 0));
   end Set_Handler_HWTime;

   ----------------------
   --  Cancel_Handler  --
   ----------------------
   procedure Cancel_Handler (Event     : in out Base_Timing_Event;
                             Cancelled : out Boolean) is
      Ret : Interfaces.C.int;
   begin
      if MaRTE.Configuration_Parameters.Not_Ada_User_Code'First then
         pragma Debug (DBG.Assert (False));
         return;
      end if;

      --  Disable event
      Ret := THO.MaRTE_Timed_Handler_Disable
        (Event.Timed_Handler'Unrestricted_Access);
      pragma Debug (DBG.Assert (Ret = 0));

      Cancelled := Event.Handler /= null;
      Event.Handler := null;
   end Cancel_Handler;

end MaRTE.Kernel.Timed_Handlers.Ada_Timing_Events;
