------------------------------------------------------------------------------
-- --------------------        M a R T E   O S          ------------------- --
------------------------------------------------------------------------------
--                                                             V1.8     080922
--
--           'A d a . A p p l i c a t i o n _ S c h e d u l i n g'
--
--                                  Spec
--
--
--  File 'a-appsch.ads'                                                By MAR
--                                                                    and MGH
--
--
--  Application-Defined Scheduling interface.
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2000-2008, Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael Gonzalez Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
------------------------------------------------------------------------------

with Ada.Task_Identification;
with Ada.Real_Time;
with Ada.Real_Time.Execution_Time;

with System;

with MaRTE.POSIX_Constants;
with MaRTE.Kernel.Application_Scheduling_Data;

package Ada.Application_Scheduling is

   package K renames MaRTE.Kernel;

   ----------------------
   --  Type 'Urgency'  --
   ----------------------
   type Urgency is private;
   function To_Urgency (N : Natural) return Urgency;
   pragma Inline (To_Urgency);

   --------------------------
   --  Scheduling Actions  --
   --------------------------

   type Scheduling_Actions is private;

   procedure Add_Accept
     (Sched_Actions : in out Scheduling_Actions;
      Tid : in Ada.Task_Identification.Task_Id);
   procedure Add_Reject
     (Sched_Actions : in out Scheduling_Actions;
      Tid : in Ada.Task_Identification.Task_Id);
   procedure Add_Ready
     (Sched_Actions : in out Scheduling_Actions;
      Tid : in Ada.Task_Identification.Task_Id);
   procedure Add_Ready
     (Sched_Actions : in out Scheduling_Actions;
      Tid : in Ada.Task_Identification.Task_Id;
      Urg : in Urgency);
   procedure Add_Timed_Task_Activation
     (Sched_Actions : in out Scheduling_Actions;
      Tid : in Ada.Task_Identification.Task_Id;
      Urg : in Urgency;
      At_Time : in Ada.Real_Time.Time);
   procedure Add_Suspend
     (Sched_Actions : in out Scheduling_Actions;
      Tid : in Ada.Task_Identification.Task_Id);
   procedure Add_Timeout
     (Sched_Actions : in out Scheduling_Actions;
      At_Time : in Ada.Real_Time.Time);
   procedure Add_Timed_Task_Notification
     (Sched_Actions : in out Scheduling_Actions;
      Tid : in Ada.Task_Identification.Task_Id;
      At_Time : in Ada.Real_Time.Time);
   procedure Add_Timer_Expiration
     (Sched_Actions : in out Scheduling_Actions;
      T             : in out Ada.Real_Time.Execution_Time.Timer;
      Abs_Time      : in     Ada.Real_Time.Execution_Time.CPU_Time);
   procedure Add_Timer_Expiration
     (Sched_Actions : in out Scheduling_Actions;
      T             : in out Ada.Real_Time.Execution_Time.Timer;
      Interval      : in     Ada.Real_Time.Time_Span);

   Scheduling_Actions_Full : exception;
   --  Raised when trying to add an scheduling action in a
   --  'Scheduling_Actions' object that has reached its maximum
   --  capacity. Never raised by 'Add_Timeout'.

   -------------------------------------
   --  Explicit Scheduler Invocation  --
   -------------------------------------

   type Message_To_Scheduler is abstract tagged
     null record;
   type Reply_From_Scheduler is abstract tagged
     null record;

   procedure Invoke
     (Msg   : access Message_To_Scheduler'Class);
   --  callable from application-scheduled task
   procedure Invoke
     (Msg   : access Message_To_Scheduler'Class;
      Reply : access Reply_From_Scheduler'Class);
   --  callable from application-scheduled task

   Task_Not_Application_Scheduled : exception;
   --  may be raised by Invoke

   ----------------------
   --  Scheduler type  --
   ----------------------

   type Scheduler is abstract tagged null record;
   type Error_Cause is
    (SRP_Rule_Violation, Invalid_Action_For_Task);

   procedure Init (Sched : out Scheduler)
      is abstract;
   procedure New_Task
     (Sched   : in out Scheduler;
      Tid     : in     Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions);
   procedure Terminate_Task
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions);
   procedure Ready
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions);
   procedure Block
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions);
   procedure Yield
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions);
   procedure Abort_Task
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions);
   procedure Change_Sched_Param
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions);
   procedure Explicit_Call
     (Sched   : in out Scheduler;
      Tid     : in     Ada.Task_Identification.Task_Id;
      Msg     : access Message_To_Scheduler'Class;
      Reply   : access Reply_From_Scheduler'Class;
      Actions : in out Scheduling_Actions);
   procedure Task_Notification
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      Actions : in out Scheduling_Actions);
   procedure Timeout
     (Sched : in out Scheduler;
      Actions : in out Scheduling_Actions);
   procedure Execution_Timer_Expiration
     (Sched         : in out Scheduler;
      Expired_Timer : in out Ada.Real_Time.Execution_Time.Timer;
      Actions       : in out Scheduling_Actions);
   procedure Priority_Inherit
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      From_Tid : in Ada.Task_Identification.Task_Id;
      Inherited_Prio : in  System.Any_Priority;
      Actions : in out Scheduling_Actions);
   procedure Priority_Uninherit
     (Sched : in out Scheduler;
      Tid : in  Ada.Task_Identification.Task_Id;
      From_Tid : in Ada.Task_Identification.Task_Id;
      Uninherited_Prio : in  System.Any_Priority;
      Actions : in out Scheduling_Actions);
   procedure Error
     (Sched : in out Scheduler;
      Tid : in Ada.Task_Identification.Task_Id;
      Cause : in  Error_Cause;
      Actions : in out Scheduling_Actions)
      is abstract;
   --  Non-abstract operations have a null body

   -------------------
   --  Event masks  --
   -------------------

   type Event_Code is
     (New_Task,  Terminate_Task, Ready, Block,
      Yield, Change_Sched_Param, Explicit_Call, Timeout,
      Priority_Inherit, Priority_Uninherit,
      Task_Notification,
      Execution_Timer_Expiration, Group_Timer_Expiration,
      Abort_Task);

   type Event_Mask is private;

   procedure Fill     (Mask : in out Event_Mask);
   procedure Empty    (Mask : in out Event_Mask);
   procedure Add      (Event : in Event_Code;
                       Mask : in out Event_Mask);
   procedure Delete   (Event : in Event_Code;
                       Mask : in out Event_Mask);
   function Is_Member (Event : in Event_Code;
                       Mask : in Event_Mask)
                      return Boolean;

   --  The event mask filters out the scheduling
   --  events that are not needed by the scheduler
   procedure Set_Event_Mask
     (Mask : in Event_Mask);
   procedure Get_Event_Mask
     (Mask : out Event_Mask);

   --  The protected event mask delays the
   --  specified scheduling events when generated
   --  while the system ceiling is larger than or
   --  equal to the specified Ceiling
   procedure Set_Protected_Event_Mask
     (Mask    : in Event_Mask;
      Ceiling : in System.Any_Priority);
   procedure Get_Protected_Event_Mask
     (Mask    : out Event_Mask;
      Ceiling : out System.Any_Priority);

   -----------------------------------------
   --  Change Task Scheduling Parameters  --
   -----------------------------------------

   type Scheduling_Parameters is
     abstract tagged null record;

   procedure Set_Parameters
     (Tid   : in Ada.Task_Identification.Task_Id;
      Param : in Scheduling_Parameters'Class);
   --  callable from any task
   --  This is a dispatching point

   procedure Get_Parameters
     (Tid   : in     Ada.Task_Identification.Task_Id;
      Param : access Scheduling_Parameters'Class);
   --  callable from any task

   --  Scheduling Policy is changed by
   --  setting the priority

private
   type Urgency is new K.Task_Urgency;

   type Scheduling_Actions is
     new K.Application_Scheduling_Data.Sched_Actions_Set;
--     type Scheduling_Actions is record
--        MaRTE_Actions    : aliased
--          K.Application_Scheduling_Data.Sched_Actions_Set;
--        Timeout_Active   : Boolean;
--        Timeout          : Ada.Real_Time.Time;

--        CPU_Timers_Count : Natural;
--     end record;

   --  type Scheduler is abstract tagged null record;

   type Event_Mask is record
      Mask : aliased K.Application_Scheduling_Data.Event_Set;
   end record;

   package PC renames MaRTE.POSIX_Constants;

   for Event_Code use
     (New_Task                   => PC.APPSCHED_NEW,
      Terminate_Task             => PC.APPSCHED_TERMINATE,
      Ready                      => PC.APPSCHED_READY,
      Block                      => PC.APPSCHED_BLOCK,
      Yield                      => PC.APPSCHED_YIELD,
      Change_Sched_Param         => PC.APPSCHED_CHANGE_SCHED_PARAM,
      Explicit_Call              => PC.APPSCHED_EXPLICIT_CALL_WITH_DATA,
      Timeout                    => PC.APPSCHED_TIMEOUT,
      Priority_Inherit           => PC.APPSCHED_PRIORITY_INHERIT,
      Priority_Uninherit         => PC.APPSCHED_PRIORITY_UNINHERIT,
      Task_Notification          => PC.APPSCHED_TASK_NOTIFICATION,
      Execution_Timer_Expiration => PC.APPSCHED_EXECUTION_TIMER_EXPIRATION,
      Group_Timer_Expiration     => PC.APPSCHED_GROUP_TIMER_EXPIRATION,
      Abort_Task                 => PC.APPSCHED_ABORT_TASK);
   for Event_Code'Size
     use K.Application_Scheduling_Data.Event_Code_T'Size;

end Ada.Application_Scheduling;
