------------------------------------------------------------------------------
-- --------------------        M a R T E   O S          ------------------- --
------------------------------------------------------------------------------
--                                                             V1.8     080922
--
--    'A d a . A p p l i c a t i o n _ S c h e d u l i n g . S u p p o r t'
--
--                                  Body
--
--
--  File 'a-apscsu.adb'                                                By MAR
--                                                                    and MGH
--
--
--  Suport for runtime and compiler implementation of
--  Application-Defined Scheduling.
--
--  ----------------------------------------------------------------------
--   Copyright (C) 2000-2008, Universidad de Cantabria, SPAIN
--
--   MaRTE OS web page: http://marte.unican.es
--   Contact Addresses: Mario Aldea Rivas          aldeam@unican.es
--                      Michael Gonzalez Harbour      mgh@unican.es
--
--  MaRTE OS  is free software; you can  redistribute it and/or  modify it
--  under the terms of the GNU General Public License  as published by the
--  Free Software Foundation;  either  version 2, or (at  your option) any
--  later version.
--
--  MaRTE OS  is distributed  in the  hope  that  it will be   useful, but
--  WITHOUT  ANY  WARRANTY;     without  even the   implied   warranty  of
--  MERCHANTABILITY  or  FITNESS FOR A  PARTICULAR PURPOSE.    See the GNU
--  General Public License for more details.
--
--  You should have received  a  copy of  the  GNU General Public  License
--  distributed with MaRTE  OS;  see file COPYING.   If not,  write to the
--  Free Software  Foundation,  59 Temple Place  -  Suite 330,  Boston, MA
--  02111-1307, USA.
--
--  As a  special exception, if you  link this  unit  with other  files to
--  produce an   executable,   this unit  does  not  by  itself cause  the
--  resulting executable to be covered by the  GNU General Public License.
--  This exception does  not however invalidate  any other reasons why the
--  executable file might be covered by the GNU Public License.
--
------------------------------------------------------------------------------
with System.OS_Interface; --  To use 'Self_Key'
with System.Task_Primitives.Operations;
with System.Tasking;
with Ada.Unchecked_Conversion;

with Ada.Real_Time.Execution_Time;

with MaRTE.Integer_Types; use MaRTE.Integer_Types;
with MaRTE.Kernel.Signals.POSIX_Functions;
with MaRTE.Kernel.Application_Scheduling_Data;
with MaRTE.Kernel.Tasks_Operations;
with MaRTE.Kernel.Tasks_Operations.Application_Scheduler;
with MaRTE.Kernel.Scheduler;

package body Ada.Application_Scheduling.Support is

   package APPSCHD_DATA renames K.Application_Scheduling_Data;
   package TO           renames K.Tasks_Operations;
   package TO_APPSCHD   renames K.Tasks_Operations.Application_Scheduler;
   package SGNL_POSIX   renames K.Signals.POSIX_Functions;
   use type K.Task_Id, MaRTE.Integer_Types.Int;

   ----------------------------------------------
   -- Conversions between 'K.Task_Id' and --
   -- 'Ada.Task_Identification.Task_Id'        --
   ----------------------------------------------
   --
   function To_MaRTE_Task_Id (Ada_Task_Id : Ada.Task_Identification.Task_Id)
                             return K.Task_Id;
   function To_MaRTE_Task_Id (Ada_Task_Id : Ada.Task_Identification.Task_Id)
                             return K.Task_Id is

      function Pthread_T_To_MaRTE_Task_Id is new Ada.Unchecked_Conversion
        (System.OS_Interface.pthread_t, K.Task_Id);
      function Convert_Ids is new Ada.Unchecked_Conversion
        (Ada.Task_Identification.Task_Id, System.Tasking.Task_Id);
   begin
      return
        Pthread_T_To_MaRTE_Task_Id
        (System.Task_Primitives.Operations.Get_Thread_Id
         (Convert_Ids (Ada_Task_Id)));
   end To_MaRTE_Task_Id;
   pragma Inline (To_MaRTE_Task_Id);

   function To_Ada_Task_Id (Kernel_Task_Id : K.Task_Id)
                           return Ada.Task_Identification.Task_Id;
   function To_Ada_Task_Id (Kernel_Task_Id : K.Task_Id)
                           return Ada.Task_Identification.Task_Id is
      function Address_To_Ada_Task_Id is
        new Unchecked_Conversion (System.Address,
                                  Ada.Task_Identification.Task_Id);
   begin
      return Address_To_Ada_Task_Id
        (Kernel_Task_Id.Specific_Data (System.OS_Interface.Self_Key).Data);
   end To_Ada_Task_Id;
   pragma Inline (To_Ada_Task_Id);

   ---------------
   --  Check_NZ --
   ---------------
   procedure Check_NZ (Result : Int);
   procedure Check_NZ (Result : Int) is
   begin
      if Result /= 0 then
         raise MaRTE_AppSched_Error;
      end if;
   end Check_NZ;

   ----------------------------------------------------------------------------
   --  Types 'Scheduler_Ac' and 'Data_For_Scheduler_Ops'  ---------------------
   ----------------------------------------------------------------------------
   type Scheduler_Ac is access all Ada.Application_Scheduling.Scheduler'Class;
   type Data_For_Scheduler_Ops is record
     Sched         : Scheduler_Ac;
     Signo         : K.Signals.Signal;
   end record;
   type Data_For_Scheduler_Ops_Ac is access all Data_For_Scheduler_Ops;

   ----------------------------------------------------------------------------
   --  Associations between schedulers and priority levels  -------------------
   ----------------------------------------------------------------------------
   type Scheduler_Associaton is record
      Sched_Task_Id : K.Task_Id;
      Sched_Data    : Data_For_Scheduler_Ops_Ac;
      Low_Prio      : System.Any_Priority;
      Installed     : Boolean;
   end record;
   Associated_Schedulers : array (System.Any_Priority) of Scheduler_Associaton
     := (others => (Sched_Task_Id => null,
                    Sched_Data    => null,
                    Low_Prio      => System.Any_Priority'First,
                    Installed     => False));

   ----------------------------------------------------------------------------
   --  Install Schedulers  ----------------------------------------------------
   ----------------------------------------------------------------------------

   ----------------------------
   --  Scheduler operations  --
   ----------------------------
   --
   --  Same parameters than the operations defined in
   --  'K.application_scheduling_data'.
   function Address_To_Data_Ac is
      new Ada.Unchecked_Conversion (System.Address, Data_For_Scheduler_Ops_Ac);

   procedure Init (Sched_Data : in System.Address);
   procedure Init (Sched_Data : in System.Address) is
      Ret : MaRTE.Integer_Types.Int;
      Set : aliased K.Signals.Signal_Set;
   begin
      Init (Address_To_Data_Ac (Sched_Data).Sched.all);
      --  Set signals waited
      Ret := K.Signals.POSIX_Functions.Sig_Empty_Set (Set'Access);
      pragma Assert (Ret = 0);
      Ret :=
        K.Signals.POSIX_Functions.Sig_Add_Set
        (Set'Access, Address_To_Data_Ac (Sched_Data).Signo);
      pragma Assert (Ret = 0);
      Ret := TO_APPSCHD.POSIX_Appschedattr_Setwaitsignalset (Set);
      pragma Assert (Ret = 0);
   end Init;

   procedure New_Task (Sched_Data : in System.Address;
                       Tid     : in K.Task_Id;
                       Actions : in APPSCHD_DATA.Sched_Actions_Set_Ac);
   procedure New_Task (Sched_Data : in System.Address;
                       Tid     : in K.Task_Id;
                       Actions : in APPSCHD_DATA.Sched_Actions_Set_Ac) is
   begin
      New_Task (Address_To_Data_Ac (Sched_Data).Sched.all,
                To_Ada_Task_Id (Tid),
                Ada.Application_Scheduling.Scheduling_Actions (Actions.all));
   end New_Task;

   procedure Ready (Sched_Data : in System.Address;
                    Tid     : in K.Task_Id;
                    Actions : in APPSCHD_DATA.Sched_Actions_Set_Ac);
   procedure Ready (Sched_Data : in System.Address;
                    Tid     : in K.Task_Id;
                    Actions : in APPSCHD_DATA.Sched_Actions_Set_Ac) is
   begin
      Ready (Address_To_Data_Ac (Sched_Data).Sched.all,
             To_Ada_Task_Id (Tid),
             Ada.Application_Scheduling.Scheduling_Actions (Actions.all));
   end Ready;

   procedure Explicit_Call
     (Sched_Data      : in System.Address;
      Tid             : in K.Task_Id;
      User_Event_Code : in Int;
      Actions         : in APPSCHD_DATA.Sched_Actions_Set_Ac);
   procedure Explicit_Call
     (Sched_Data          : in System.Address;
      Tid             : in K.Task_Id;
      User_Event_Code : in Int;
      Actions         : in APPSCHD_DATA.Sched_Actions_Set_Ac) is
   begin
      pragma Unreferenced (User_Event_Code);
      Explicit_Call
        (Address_To_Data_Ac (Sched_Data).Sched.all,
         To_Ada_Task_Id (Tid),
         null, null,
         Ada.Application_Scheduling.Scheduling_Actions (Actions.all));
   end Explicit_Call;

   procedure Task_Notification
     (Sched_Data : in System.Address;
      Tid        : in  K.Task_Id;
      Actions    : in APPSCHD_DATA.Sched_Actions_Set_Ac);
   procedure Task_Notification
     (Sched_Data : in System.Address;
      Tid        : in  K.Task_Id;
      Actions    : in APPSCHD_DATA.Sched_Actions_Set_Ac) is
   begin
      Task_Notification
        (Address_To_Data_Ac (Sched_Data).Sched.all,
         To_Ada_Task_Id (Tid),
         Ada.Application_Scheduling.Scheduling_Actions (Actions.all));
   end Task_Notification;

   procedure Timeout (Sched_Data : in System.Address;
                      Actions    : in APPSCHD_DATA.Sched_Actions_Set_Ac);
   procedure Timeout (Sched_Data : in System.Address;
                      Actions    : in APPSCHD_DATA.Sched_Actions_Set_Ac) is
   begin
      Timeout (Address_To_Data_Ac (Sched_Data).Sched.all,
               Ada.Application_Scheduling.Scheduling_Actions (Actions.all));
   end Timeout;

   procedure Signal (Sched_Data : in System.Address;
                     Siginfo   : in K.Signals.Siginfo_T;
                     Actions    : in APPSCHD_DATA.Sched_Actions_Set_Ac);
   procedure Signal (Sched_Data : in System.Address;
                     Siginfo    : in K.Signals.Siginfo_T;
                     Actions    : in APPSCHD_DATA.Sched_Actions_Set_Ac) is
      type Execution_Time_Timer_Ac is
        access all Ada.Real_Time.Execution_Time.Timer;
      function To_Execution_Time_Timer_Ac is
         new Ada.Unchecked_Conversion (MaRTE.Integer_Types.Int,
                                       Execution_Time_Timer_Ac);
   begin
      Execution_Timer_Expiration
        (Address_To_Data_Ac (Sched_Data).Sched.all,
         To_Execution_Time_Timer_Ac (Siginfo.Value).all,
         Ada.Application_Scheduling.Scheduling_Actions (Actions.all));
   end Signal;

   -------------------------
   --  Install_Scheduler  --
   -------------------------
   procedure Install_Scheduler
     (Sched      : access Ada.Application_Scheduling.Scheduler'Class;
      High_Prio  : in     System.Any_Priority;
      Low_Prio   : in     System.Any_Priority) is
      T : aliased K.Task_Id;
      Scheduler_Ops : aliased APPSCHD_DATA.Scheduler_Operations :=
        (Init              => Init'Access,
         New_Task          => New_Task'Access,
         Ready             => Ready'Access,
         Explicit_Call     => Explicit_Call'Access,
         Task_Notification => Task_Notification'Access,
         Timeout           => Timeout'Access,
         Signal            => Signal'Access);
      use type K.Signals.Signal;
      Sched_Data    : Data_For_Scheduler_Ops_Ac;
   begin
      Sched_Data :=
        new Data_For_Scheduler_Ops'
        (Sched => Scheduler_Ac (Sched),
         Signo => K.Signals.Realtime_Signal'First +
         K.Signals.Signal (High_Prio));
      if TO_APPSCHD.Posix_Appsched_Scheduler_Create
        (Scheduler_Ops'Access, Sched_Data.all'Address,
         K.Task_Priority (High_Prio), T'Access) /= 0 then
         --  Impossible create application scheduler task
         raise Reached_Maximum_Number_Of_Application_Schedulers;
      end if;

      --  Associate with priority level
      Associated_Schedulers (Low_Prio .. High_Prio) :=
        (others => (Sched_Task_Id => T,
                    Sched_Data    => Sched_Data,
                    Installed     => True,
                    Low_Prio      => Low_Prio));
   end Install_Scheduler;

   ----------------------------------------------------------------------------
   --  Manage Schedulers and Appscheduled Tasks  ------------------------------
   ----------------------------------------------------------------------------

   ---------------------------
   --  Scheduler_Installed  --
   ---------------------------
   function Scheduler_Installed (Prio : in System.Any_Priority)
                                return Boolean is
   begin
      return Associated_Schedulers (Prio).Installed;
   end Scheduler_Installed;

   --------------------------------
   --  Attach_Task_To_Scheduler  --
   --------------------------------
   procedure Attach_Task_To_Scheduler
     (Tid  : in Ada.Task_Identification.Task_Id;
      Prio : in System.Any_Priority) is
      Sched_Param : aliased TO.Sched_Param;
      Policy      : aliased TO.Scheduling_Policies;
   begin
      if not Associated_Schedulers (Prio).Installed then
         raise Not_Scheduler_At_Priority;
      end if;

      --  Set appscheduler attribute
      Check_NZ (TO_APPSCHD.Pthread_Setappscheduler
                                 (To_MaRTE_Task_Id (Tid),
                                  Associated_Schedulers (Prio).Sched_Task_Id));

      --  Get current scheduling parameters
      Check_NZ (TO.Pthread_Getschedparam (To_MaRTE_Task_Id (Tid),
                                          Policy'Access,
                                          Sched_Param'Access));

      --  Set PL as the base priority of task
      K.Scheduler.Self.Active_Preemption_Level :=
        K.Task_Preemption_Level (Sched_Param.Sched_Priority);
      K.Scheduler.Self.Base_Preemption_Level :=
        K.Task_Preemption_Level (Sched_Param.Sched_Priority);

      --  The priority of the task is the lower priorty in the range
      --  associated to its scheduler
      Sched_Param.Sched_Priority :=
        K.Task_Priority (Associated_Schedulers (Prio).Low_Prio);

      --  Set new policy and priority
      Check_NZ (TO.Pthread_Setschedparam (To_MaRTE_Task_Id (Tid),
                                          K.SCHED_APP,
                                          Sched_Param'Access));
   end Attach_Task_To_Scheduler;

   ----------------------------------
   --  Detach_Task_From_Scheduler  --
   ----------------------------------
   procedure Detach_Task_From_Scheduler
     (Tid  : in Ada.Task_Identification.Task_Id;
      Prio : in System.Any_Priority) is
      Sched_Param : aliased TO.Sched_Param;
      Policy      : aliased TO.Scheduling_Policies;
      Sched       : aliased K.Task_Id;
   begin
      if not Associated_Schedulers (Prio).Installed then
         raise Not_Scheduler_At_Priority;
      end if;

      --  Get appscheduler attribute
      Check_NZ (TO_APPSCHD.Pthread_Getappscheduler (To_MaRTE_Task_Id (Tid),
                                                    Sched'Access));
      if Sched /= Associated_Schedulers (Prio).Sched_Task_Id then
         raise Task_Not_Attached_To_Scheduler;
      end if;

      --  Get current scheduling parameters
      Check_NZ (TO.Pthread_Getschedparam (To_MaRTE_Task_Id (Tid),
                                          Policy'Access,
                                          Sched_Param'Access));
      --  Change policy
      Check_NZ (TO.Pthread_Setschedparam (To_MaRTE_Task_Id (Tid),
                                          K.SCHED_FIFO,
                                          Sched_Param'Access));
   end Detach_Task_From_Scheduler;

   ----------------------------
   --  Generate Abort Event  --
   ----------------------------
   procedure Generate_Abort_Event_For_Scheduler
     (Aborted_Task : in Ada.Task_Identification.Task_Id;
      Prio         : in System.Any_Priority) is
      function Ada_Task_Id_To_Int is
         new Unchecked_Conversion (Ada.Task_Identification.Task_Id, Int);
      Sched : aliased K.Task_Id;
   begin
      if not Associated_Schedulers (Prio).Installed then
         raise Not_Scheduler_At_Priority;
      end if;

      --  Get appscheduler attribute
      Check_NZ (TO_APPSCHD.Pthread_Getappscheduler
                                           (To_MaRTE_Task_Id (Aborted_Task),
                                            Sched'Access));
      if Sched /= Associated_Schedulers (Prio).Sched_Task_Id then
         raise Task_Not_Attached_To_Scheduler;
      end if;

      if SGNL_POSIX.Sigqueue (0, Associated_Schedulers (Prio).Sched_Data.Signo,
                              Ada_Task_Id_To_Int (Aborted_Task)) /= 0 then
         raise Not_Enough_Resources_To_Send_Signal;
      end if;
   end Generate_Abort_Event_For_Scheduler;

end Ada.Application_Scheduling.Support;
