------------------------------------------------------------------------------
--                                                                          --
--                         GNAT RUN-TIME COMPONENTS                         --
--                                                                          --
--      A D A . E X E C U T I O N _ T I M E . G R O U P _ B U D G E T S     --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
-- This specification is derived from the Ada Reference Manual for use with --
-- GNAT.  In accordance with the copyright of that document, you can freely --
-- copy and modify this specification,  provided that if you redistribute a --
-- modified version,  any changes that you have made are clearly indicated. --
--                                                                          --
------------------------------------------------------------------------------

--  MaRTE OS version

with Interfaces.C;
with Ada.Unchecked_Conversion;
with Ada.Real_Time;
with System.Tasking;
with System.Task_Primitives.Operations;

with MaRTE.Kernel.Debug;
with MaRTE.HAL;
with MaRTE.Kernel.Task_Sets.Operations;
with MaRTE.Kernel.Timers.Internals;
with MaRTE.Kernel.Timed_Handlers.Operations;

package body Ada.Execution_Time.Group_Budgets is

   package DBG renames MaRTE.Kernel.Debug;
   package HAL renames MaRTE.HAL;
   package TSO renames MaRTE.Kernel.Task_Sets.Operations;
   package TMRI renames MaRTE.Kernel.Timers.Internals;
   package THO renames MaRTE.Kernel.Timed_Handlers.Operations;

   use type Interfaces.C.int;
   use type Ada.Real_Time.Time_Span;

   -----------------------
   -- Local Subprograms --
   -----------------------

   function To_Group_Budget_Handler is new Ada.Unchecked_Conversion
     (THAda.Base_Timing_Event_Handler, Group_Budget_Handler);

   function To_Ada_TE_Handler is new Ada.Unchecked_Conversion
     (Group_Budget_Handler, THAda.Base_Timing_Event_Handler);

   function To_MaRTE_Task_Id (T : Ada.Task_Identification.Task_Id)
                              return MaRTE.Kernel.Task_Id;
   function To_MaRTE_Task_Id (T : Ada.Task_Identification.Task_Id)
                              return MaRTE.Kernel.Task_Id is
      function UC is new Ada.Unchecked_Conversion
        (Ada.Task_Identification.Task_Id, System.Tasking.Task_Id);
   begin
      return MaRTE.Kernel.Task_Id
           (System.Task_Primitives.Operations.Get_Thread_Id (UC (T)));
   end To_MaRTE_Task_Id;

   ----------------
   -- Initialize --
   ----------------

   procedure Initialize (GB : in out Group_Budget) is
      Ret : Interfaces.C.int;
   begin
      --  Create thread set

      Ret := TSO.Marte_Thread_Set_Create (GB.Set'Access);
      pragma Debug (DBG.Assert (Ret = 0));

      --  Get group clock

      Ret := TSO.Marte_Getgroupclockid (GB.Set, GB.Base_Clock'Access);
      pragma Debug (DBG.Assert (Ret = 0));

      --  Initialize timed handler

      GB.Based_On_Group_Clock := True;
      THAda.Initialize (THAda.Base_Timing_Event (GB));
   end Initialize;

   --------------
   -- Finalize --
   --------------

   procedure Finalize (GB : in out Group_Budget) is
      Ret : Interfaces.C.int;
   begin
      --  Marte_Thread_Set_Destroy removes tasks from the set and destroyes the
      --  associated timed handler (so it isn't necessary to call
      --  THAda.Finalize)
      Ret := TSO.Marte_Thread_Set_Destroy (GB.Set);
      pragma Debug (DBG.Assert (Ret = 0));
   end Finalize;

   --------------
   -- Add_Task --
   --------------

   procedure Add_Task
     (GB : in out Group_Budget;
      T  : Ada.Task_Identification.Task_Id) is
      Ret : Interfaces.C.int;
   begin
      Ret := TSO.Marte_Thread_Set_Add_Thread (GB.Set, To_MaRTE_Task_Id (T));
      if Ret /= 0 then
         raise Group_Budget_Error;
      end if;
   end Add_Task;

   -----------------
   -- Remove_Task --
   -----------------

   procedure Remove_Task
     (GB : in out Group_Budget;
      T  : Ada.Task_Identification.Task_Id) is
      Ret : Interfaces.C.int;
   begin
      Ret := TSO.Marte_Thread_Set_Del_Thread (GB.Set, To_MaRTE_Task_Id (T));
      if Ret /= 0 then
         raise Group_Budget_Error;
      end if;
   end Remove_Task;

   ---------------
   -- Is_Member --
   ---------------

   function Is_Member
     (GB : Group_Budget;
      T  : Ada.Task_Identification.Task_Id) return Boolean is
      Ret : Interfaces.C.int;
      Is_Member : aliased Interfaces.C.int;
   begin
      Ret := TSO.Marte_Thread_Set_Is_Member (GB.Set, To_MaRTE_Task_Id (T),
                                             Is_Member'Access);
      pragma Debug (DBG.Assert (Ret = 0));
      return Is_Member = 1;
   end Is_Member;

   -----------------------
   -- Is_A_Group_Member --
   -----------------------

   function Is_A_Group_Member
     (T : Ada.Task_Identification.Task_Id) return Boolean is
      Ret : Interfaces.C.int;
      Is_Member : aliased Interfaces.C.int;
   begin
      Ret := TSO.Marte_Thread_Set_Is_A_Set_Member (To_MaRTE_Task_Id (T),
                                                   Is_Member'Access);
      pragma Debug (DBG.Assert (Ret = 0));
      return Is_Member = 1;
   end Is_A_Group_Member;

   -------------
   -- Members --
   -------------

   function Members (GB : Group_Budget) return Task_Array is
      pragma Unreferenced (GB);
      Task_Array_Ac : constant access Task_Array := new Task_Array (1 .. 1);
   begin
      return Task_Array_Ac.all;
   end Members;

   ---------------
   -- Replenish --
   ---------------

   procedure Replenish
     (GB : in out Group_Budget;
      To : Ada.Real_Time.Time_Span) is
      use type Ada.Real_Time.Time_Span;
   begin
      if To < Ada.Real_Time.Time_Span_Zero then
         raise Group_Budget_Error;
      end if;
      GB.T := To;
      THAda.Set_Handler_HWTime
        (THAda.Base_Timing_Event (GB),
         HAL.Duration_To_HWTime (Ada.Real_Time.To_Duration (To)),
         GB.Handler,
         MaRTE.Kernel.Timed_Handlers.ABSOLUTE_TIMER);
   end Replenish;

   ---------
   -- Add --
   ---------

   procedure Add
     (GB       : in out Group_Budget;
      Interval : Ada.Real_Time.Time_Span) is
      use type Ada.Real_Time.Time_Span;
   begin
      GB.T := GB.T + Interval;

      --  Never below zero RM D.14.2(22)
      if GB.T < Ada.Real_Time.Time_Span_Zero then
         GB.T := Ada.Real_Time.Time_Span_Zero;
      end if;

      THAda.Set_Handler_HWTime
        (THAda.Base_Timing_Event (GB),
         HAL.Duration_To_HWTime (Ada.Real_Time.To_Duration (GB.T)),
         GB.Handler,
         MaRTE.Kernel.Timed_Handlers.ABSOLUTE_TIMER);
   end Add;

   ------------------------
   -- Budget_Has_Expired --
   ------------------------

   function Budget_Has_Expired (GB : Group_Budget) return Boolean is
      Ret : Interfaces.C.int;
      Expired : aliased Interfaces.C.int;
   begin
      Ret := THO.MaRTE_Timed_Handler_Expired
        (GB.Timed_Handler'Unrestricted_Access, Expired'Access);
      pragma Assert (Ret = 0);
      return Expired = 1;
   end Budget_Has_Expired;

   ----------------------
   -- Budget_Remaining --
   ----------------------

   function Budget_Remaining
     (GB : Group_Budget) return Ada.Real_Time.Time_Span is
   begin
      return Ada.Real_Time.To_Time_Span
        (HAL.HWTime_To_Duration
           (TMRI.Time_Remaining_To_CPU_Time_Event
              (GB.Timed_Handler'Unrestricted_Access,
               HAL.Get_HWTime)));
   end Budget_Remaining;

   -----------------
   -- Set_Handler --
   -----------------

   procedure Set_Handler
     (GB      : in out Group_Budget;
      Handler : Group_Budget_Handler) is
   begin
      THAda.Set_Handler_HWTime
        (THAda.Base_Timing_Event (GB),
         HAL.Duration_To_HWTime (Ada.Real_Time.To_Duration (GB.T)),
         To_Ada_TE_Handler (Handler),
         MaRTE.Kernel.Timed_Handlers.ABSOLUTE_TIMER);
   end Set_Handler;

   ---------------------
   -- Current_Handler --
   ---------------------

   function Current_Handler (GB : Group_Budget) return Group_Budget_Handler is
   begin
      return To_Group_Budget_Handler (GB.Handler);
   end Current_Handler;

   --------------------
   -- Cancel_Handler --
   --------------------

   procedure Cancel_Handler
     (GB        : in out Group_Budget;
      Cancelled : out Boolean)
   is
   begin
      THAda.Cancel_Handler (THAda.Base_Timing_Event (GB), Cancelled);
   end Cancel_Handler;

end Ada.Execution_Time.Group_Budgets;
