with Ada.Unchecked_Conversion;

with MaRTE.Kernel.Application_Scheduling_Data;
with MaRTE.Kernel.Tasks_Operations.Application_Scheduler;
with MaRTE.Kernel.Signals;
with MaRTE.Kernel.Signals.POSIX_Functions;
with MaRTE.Kernel.Timers;
with MaRTE.Kernel.Scheduler;
with MaRTE.HAL;
pragma Elaborate_All (MaRTE.HAL);
with MaRTE.POSIX_Pthread;
with MaRTE.Integer_Types;
--  with Kernel_Console;

package body System.Round_Robin_Scheduler is

   package K renames MaRTE.Kernel;

   package APPSCHD_DATA renames K.Application_Scheduling_Data;
   package TO           renames K.Tasks_Operations;
   package TO_APPSCHD   renames K.Tasks_Operations.Application_Scheduler;
   package SCHD         renames K.Scheduler;
   package HAL          renames MaRTE.HAL;
   use type MaRTE.Integer_Types.Int, K.Signals.Signal;

   Default_Quantum : constant HAL.HWTime := HAL.Duration_To_HWTime (0.01);

   ---------------------------------------------------------------------------
   --  Round Robin scheduler  ------------------------------------------------
   ---------------------------------------------------------------------------

   ----------------------
   --  Scheduler Data  --
   ----------------------
   type Quantum_Array is array (System.Any_Priority) of HAL.HWTime;

   type RR_Scheduler_Data is record
      Quantum         : Quantum_Array;
      Signo           : K.Signals.Signal;
      Key             : MaRTE.POSIX_Pthread.Pthread_Key_T;
      RR_In_Partition : Boolean;
   end record;

   --  type RR_Scheduler_Data_Ac is access all RR_Scheduler_Data;
   --  function To_RR_Scheduler_Data is
   --     new Ada.Unchecked_Conversion (System.Address, RR_Scheduler_Data_Ac);

   Schr_Data : aliased RR_Scheduler_Data;

   Schr_Task_Id : aliased K.Task_Id;

   Initialized : Boolean := False;
   --  Only should exist one RR scheduler

   -----------------
   --  Task Data  --
   -----------------
   type RR_Task_Data is record
      Prio            : System.Any_Priority;
      Scheduled_By_RR : Boolean;
      Timer           : K.Timers.Timer_Id;
   end record;
   type RR_Task_Data_Ac is access RR_Task_Data;
   function To_Address is
      new Ada.Unchecked_Conversion (RR_Task_Data_Ac, System.Address);
   function To_RR_Task_Data_Ac is
      new Ada.Unchecked_Conversion (System.Address, RR_Task_Data_Ac);

   --  type Action_Str is new String (1 .. 4);
   --  procedure Display_Task_Data (Act : in Action_Str;
   --                             Tid : in K.Task_Id;
   --                             T   : in RR_Task_Data_Ac);

   ------------
   --  Init  --
   ------------
   procedure Schr_RR_Init (Sched_Data : in System.Address);
   procedure Schr_RR_Init (Sched_Data : in System.Address) is
      pragma Unreferenced (Sched_Data);
      Ret : MaRTE.Integer_Types.Int;
      Set : aliased K.Signals.Signal_Set;
      Mask : aliased APPSCHD_DATA.Event_Set;
   begin
      --  Set signals waited
      Ret := K.Signals.POSIX_Functions.Sig_Empty_Set (Set'Access);
      pragma Assert (Ret = 0);
      Ret := K.Signals.POSIX_Functions.Sig_Add_Set (Set'Access,
                                                         Schr_Data.Signo);
      pragma Assert (Ret = 0);
      Ret := TO_APPSCHD.POSIX_Appschedattr_Setwaitsignalset (Set);
      pragma Assert (Ret = 0);

      --  Set event mask
      Ret := TO_APPSCHD.POSIX_Appsched_Fillset (Mask'Access);
      pragma Assert (Ret = 0);
      --  New_Task
      Ret := TO_APPSCHD.POSIX_Appsched_Delset (Mask'Access,
                                               APPSCHD_DATA.APPSCHED_NEW);
      pragma Assert (Ret = 0);
      --  Ready
      Ret := TO_APPSCHD.POSIX_Appsched_Delset (Mask'Access,
                                               APPSCHD_DATA.APPSCHED_READY);
      pragma Assert (Ret = 0);
      --  Signal
      Ret := TO_APPSCHD.POSIX_Appsched_Delset (Mask'Access,
                                               APPSCHD_DATA.APPSCHED_SIGNAL);
      pragma Assert (Ret = 0);
      Ret := TO_APPSCHD.POSIX_Appschedattr_Seteventmask (Mask);
      pragma Assert (Ret = 0);
   end Schr_RR_Init;

   ----------------
   --  New_Task  --
   ----------------
   procedure Schr_RR_New_Task
     (Sched_Data : in System.Address;
      Tid     : in K.Task_Id;
      Actions : in APPSCHD_DATA.Sched_Actions_Set_Ac);
   procedure Schr_RR_New_Task
     (Sched_Data : in System.Address;
      Tid     : in K.Task_Id;
      Actions : in APPSCHD_DATA.Sched_Actions_Set_Ac) is

      pragma Unreferenced (Sched_Data);
      Ret : MaRTE.Integer_Types.Int;
      Addr : aliased System.Address;
      T_Data : RR_Task_Data_Ac;
      use type K.Timers.Timer_Id;
   begin
      --  Get RR task data
      Ret :=
        MaRTE.POSIX_Pthread.Pthread_Getspecific_From (Schr_Data.Key, Tid,
                                                      Addr'Access);
      pragma Assert (Ret = 0);
      T_Data := To_RR_Task_Data_Ac (Addr);
      pragma Assert (T_Data /= null);

      --  Create CPU-time timer (if it is not created yet)
      if T_Data.Timer = null then
         declare
            Event : K.Signals.Signal_Event;
            Clock : aliased K.Timers.Clock_Id;
            function To_Int is
               new Ada.Unchecked_Conversion (K.Task_Id,
                                             MaRTE.Integer_Types.Int);
         begin
            Ret := K.Timers.Pthread_Getcpuclockid (Tid, Clock'Access);
            pragma Assert (Ret = 0);
            Event := (Event_Notification => K.Signals.SIGNAL_NOTIFICATION,
                      Event_Signal       => Schr_Data.Signo,
                      Event_Sigval       => To_Int (Tid));
            T_Data.Timer := K.Timers.Create_Timer (Clock, Event);
            pragma Assert (T_Data.Timer /= null);
         end;
      end if;

      --  Program CPU-time timer for the end of its quantum: it is
      --  armed as a relative timer with period
      --  Schr.Quantum (T_Data.Prio)
      K.Timers.Arm_Timer
        (T_Data.Timer,
         Options => K.Timers.Timer_Options'(0),  --  Relative timer
--         New_State => (Initial  => 0, Interval => 0));
         New_State => (Initial  => Schr_Data.Quantum (T_Data.Prio),
                       Interval => Schr_Data.Quantum (T_Data.Prio)));
         pragma Assert (not SCHD.Self.POSIX_Error);

      --  Accept task
      Ret := TO_APPSCHD.POSIX_Appsched_Actions_Addaccept (Actions, Tid);
      pragma Assert (Ret = 0);

      --  Activate task
      Ret := TO_APPSCHD.POSIX_Appsched_Actions_Addactivateurg
        (Actions, Tid, 0);
      pragma Assert (Ret = 0);

   end Schr_RR_New_Task;

   -------------
   --  Ready  --
   -------------
   procedure Schr_RR_Ready (Sched_Data : in System.Address;
                            Tid     : in K.Task_Id;
                            Actions : in APPSCHD_DATA.Sched_Actions_Set_Ac);
   procedure Schr_RR_Ready (Sched_Data : in System.Address;
                            Tid     : in K.Task_Id;
                            Actions : in APPSCHD_DATA.Sched_Actions_Set_Ac) is
      pragma Unreferenced (Sched_Data);
      Addr : aliased System.Address;
      T_Data : RR_Task_Data_Ac;
      Ret : MaRTE.Integer_Types.Int;
   begin
      --  Get task data
      Ret :=
        MaRTE.POSIX_Pthread.Pthread_Getspecific_From (Schr_Data.Key, Tid,
                                                      Addr'Access);
      pragma Assert (Ret = 0);
      T_Data := To_RR_Task_Data_Ac (Addr);
      pragma Assert (T_Data /= null);

      --  Program CPU-time timer for the end of its quantum: it is
      --  armed as a relative timer with period
      --  Schr.Quantum (T_Data.Prio)
      K.Timers.Arm_Timer
        (T_Data.Timer,
         Options => K.Timers.Timer_Options'(0),  --  Relative timer
         New_State => (Initial  => Schr_Data.Quantum (T_Data.Prio),
                       Interval => Schr_Data.Quantum (T_Data.Prio)));
      pragma Assert (not SCHD.Self.POSIX_Error);

      --  Activate task
      Ret := TO_APPSCHD.POSIX_Appsched_Actions_Addactivateurg
        (Actions, Tid, 0);
      pragma Assert (Ret = 0);

   end Schr_RR_Ready;

   --------------
   --  Signal  --
   --------------
   --
   --  CPU-Time timer expiration
   procedure Schr_RR_Signal (Sched_Data : in System.Address;
                             Siginfo : in K.Signals.Siginfo_T;
                             Actions : in APPSCHD_DATA.Sched_Actions_Set_Ac);
   procedure Schr_RR_Signal (Sched_Data : in System.Address;
                             Siginfo : in K.Signals.Siginfo_T;
                             Actions : in APPSCHD_DATA.Sched_Actions_Set_Ac) is
      pragma Unreferenced (Sched_Data);
      function To_Tid is
         new Ada.Unchecked_Conversion (MaRTE.Integer_Types.Int,
                                       K.Task_Id);
      Tid : constant K.Task_Id := To_Tid (Siginfo.Value);
      Ret : MaRTE.Integer_Types.Int;
   begin
      --  Move task to the tail of its priority queue
      Ret :=
        TO_APPSCHD.POSIX_Appsched_Actions_Addactivateurg (Actions, Tid, 0);
      pragma Assert (Ret = 0);

   end Schr_RR_Signal;

   ----------------------------------------------------------------------------
   --  Procedures in the interface  -------------------------------------------
   ----------------------------------------------------------------------------

   --------------
   --  Create  --
   --------------
   procedure Create (High_Prio : in System.Any_Priority) is
      Ret : MaRTE.Integer_Types.Int;
      Scheduler_Ops : aliased APPSCHD_DATA.Scheduler_Operations :=
        (Init              => Schr_RR_Init'Access,
         New_Task          => Schr_RR_New_Task'Access,
         Ready             => Schr_RR_Ready'Access,
         Explicit_Call     => null,
         Task_Notification => null,
         Timeout           => null,
         Signal            => Schr_RR_Signal'Access);
      Key : aliased MaRTE.POSIX_Pthread.Pthread_Key_T;
   begin
      pragma Assert (not Initialized);
      Initialized := True;

      --  Create pthread specific-data key
      Ret := MaRTE.POSIX_Pthread.Pthread_Key_Create (Key'Unrestricted_Access,
                                                     System.Null_Address);

      pragma Assert (Ret = 0);

      --  Init scheduler data
      Schr_Data :=
        (Quantum => (others => Default_Quantum),
         Signo   => K.Signals.Realtime_Signal'First,
         Key     => Key,
         RR_In_Partition => True);
      --  Create scheduler
      Ret := TO_APPSCHD.Posix_Appsched_Scheduler_Create
        (Scheduler_Ops'Access, Schr_Data'Address,
         K.Task_Priority (High_Prio), Schr_Task_Id'Access);
      pragma Assert (Ret = 0);
   end Create;

   --------------------------
   --  No_RR_In_Partition  --
   --------------------------
   procedure No_RR_In_Partition is
   begin
      pragma Assert (not Initialized);
      Initialized := True;

      --  Init scheduler data
      Schr_Data.RR_In_Partition := False;
   end No_RR_In_Partition;

   ---------------------
   --  Register_Task  --
   ---------------------
   --
   --  Called from System.Task_Primitives.Operations.Set_Priority
   procedure Register_Task_After_Prio_Change
     (T : in System.OS_Interface.pthread_t) is

      Sched_Param : aliased TO.Sched_Param;
      Policy      : aliased TO.Scheduling_Policies;
      Ret         : MaRTE.Integer_Types.Int;
      Addr : aliased System.Address;
      T_Data : RR_Task_Data_Ac;
      use type K.Task_Types;
   begin
      pragma Assert (Initialized);
      pragma Assert (Schr_Data.RR_In_Partition);

      --  Get RR task specific data
      Ret :=
        MaRTE.POSIX_Pthread.Pthread_Getspecific_From (Schr_Data.Key, T,
                                                      Addr'Access);
      pragma Assert (Ret = 0);
      T_Data := To_RR_Task_Data_Ac (Addr);

      --  If RR specific data is set for this task, that means the
      --  task has already started its body and, consecuently, it can
      --  be changed to policy SCHED_APP
      if T_Data /= null and then not T_Data.Scheduled_By_RR then
         --  Set appscheduler attribute
         Ret := TO_APPSCHD.Pthread_Setappscheduler (T, Schr_Task_Id);
         pragma Assert (Ret = 0);

         --  Get scheduling parameters
         Ret := TO.Pthread_Getschedparam (T, Policy'Access,
                                          Sched_Param'Access);
         pragma Assert (Ret = 0);

         --  Set SCHED_APP for the task
         Ret := TO.Pthread_Setschedparam (T, K.SCHED_APP,
                                          Sched_Param'Access);
         pragma Assert (Ret = 0);
         T_Data.Scheduled_By_RR := True;

      elsif T.Task_Type = K.MAIN_TASK then
         Register_If_Appropriate (T);
      end if;

      --  If RR specific data is not set, it will be set, and task
      --  will be changed to policy SCHED_APP, in
      --  Register_If_Appropriate

   end Register_Task_After_Prio_Change;

   -------------------------------
   --  Register_If_Appropriate  --
   -------------------------------
   --
   --  Called from System.task_states.Task_Wrapper
   procedure Register_If_Appropriate (T : in System.OS_Interface.pthread_t) is

      Sched_Param : aliased TO.Sched_Param;
      Policy      : aliased TO.Scheduling_Policies;
      Ret         : MaRTE.Integer_Types.Int;
      T_Data : RR_Task_Data_Ac;
   begin
      pragma Assert (Initialized);

      if not Schr_Data.RR_In_Partition then
         --  Nothing to to
         return;
      end if;

      --  Create RR specific data with null timer
      T_Data := new RR_Task_Data;
      T_Data.all := (Prio            => System.Any_Priority (T.Base_Prio),
                     Scheduled_By_RR => False,
                     Timer           => null);
      --  Associate task data
      Ret := MaRTE.POSIX_Pthread.Pthread_Setspecific_For (Schr_Data.Key, T,
                                                          To_Address (T_Data));
      pragma Assert (Ret = 0);

      --  If task is in a RR priority level => Change policy to SCHED_APP
      --  Set appscheduler attribute
      Ret := TO_APPSCHD.Pthread_Setappscheduler (T, Schr_Task_Id);
      pragma Assert (Ret = 0);

      --  Get current scheduling parameters
      Ret := TO.Pthread_Getschedparam (T, Policy'Access, Sched_Param'Access);
      pragma Assert (Ret = 0);

      --  Set SCHED_APP for the task
      Ret := TO.Pthread_Setschedparam (T, K.SCHED_APP,
                                       Sched_Param'Access);
      pragma Assert (Ret = 0);
      T_Data.Scheduled_By_RR := True;

   end Register_If_Appropriate;

end System.Round_Robin_Scheduler;
