#!/usr/bin/perl -w
#----------------------------------------------------------------------------
#--------------------        M a R T E     O S        -----------------------
#----------------------------------------------------------------------------
#
#                     't a s k s _ i n s p e c t o r'
#
#                             Perl Script
#
#
# File 'tasks_inspector.pl'                                          By MAR.
#

# This script translates a file with scheduling events generated by
# MaRTE OS in another set of files to be read for 'gnuplot'.

# Use:
#     $ tasks_inspector <event_file>

# Having the events file the following structure:
#
#    Empty or irrelevant lines    
#    #0#ID(12,31)#
#    #0#CF(Clock_freq,0)#
#    #130374414940538808#TR(0,0)#
#    #130374415072730860#NT(1,5)#
#    #130374415073083994#TR(1,5)#
#    #130374415073711017#NT(2,5)#
#    #130374415074310073#NT(3,5)#
#    #130374415074946417#NT(4,5)#
#    #130374415075306926#TB(0,7)#
#    #130374415075653761#CS(0,7)->(1,5)#
#    #130374415076936269#TR(2,5)#
#    ...

# The tasks and mutexes are printed with the following value depending
# on their state and priority.

#
#                         |  mutex locked -> $MUTEXES                       |
#                         |                  + (2*$PRIO_GAP*(id-$NTASKS))   | 
#                         |                  + $PRIO_GAP                    |
#                         |  free mutex -> $MUTEXES                         |
#                         |                + (2*$PRIO_GAP*(id-$NTASKS))     |
# $MUTEXES = $RUNNING     +-------------------------------------------------+
#          + $PRIO_GAP    |                                                 |
#               * $NPRIOS |                                                 |
#                         | Running tasks -> $RUNNING + prio*$PRIO_GAP + id |
# $RUNNING = $PRIO_GAP    +-------------------------------------------------+
#              * $NPRIOS  |                                                 |
#                         | Ready tasks -> $READY + prio*$PRIO_GAP + id     |
#                         |                                                 |
#                         |                                                 |
# $READY = 0              +-------------------------------------------------+
#                         |                                                 |
#                         | Blocked tasks -> $BLOCKED + id                  |
#                         |                                                 |
# $BLOCKED = -2*$PRIO_GAP +-------------------------------------------------+
#                         |                                                 |
#                         | Terminated tasks -> $BLOCKED + id               |
# $TERMINATED =           |                                                 |
#           -4*$PRIO_GAP  +-------------------------------------------------+

#---------------------------------------------------------------------------*/
$NTASKS     = 0;
$NMUTEXES   = 20; # Is it enough in your case ?
$NPRIOS     = 0;
$PRIO_GAP   = 0;
$BLOCKED    = 0;
$TERMINATED = 0;
$READY      = 0;
$RUNNING    = 0;
$MUTEXES    = 0;
$is_the_first = 1;
$firts_timestamp = 0;
$main_initial_prio = 0;
$TIME_CONVERSION = 0.0; # 

@events = ();

$#ARGV == 0 or die "use: task_inspector events_file";

$file_name = $ARGV[0];
open EVENTS_FILE, "<$file_name" or die "Can't open $file_name";
@events = <EVENTS_FILE>;

# Pick firsts events
while (1) {
    ($event, @events) = @events;
    if ($event ne "") {
	if ($event =~ /\#0\#ID\((\d+)\,(\d+)\)\#/) {
	    #           #0#ID(Num_Tasks,Num_Prio)#
	    $NTASKS  = $1 + 2; # ("+ 2" for the idle and sighandler tasks)
	    $NPRIOS  = $2;
	    $PRIO_GAP = 25 * (($NTASKS % 25) + 1); # 25 or 50 or ...
	    $BLOCKED = -2 * $PRIO_GAP;
	    $TERMINATED = -4 * $PRIO_GAP;
	    $READY   = 0;
	    $RUNNING = $PRIO_GAP * $NPRIOS;
	    $MUTEXES = $RUNNING + $PRIO_GAP * $NPRIOS;

	    # The second event gives the clock frequency
	    ($event, @events) = @events; 
	    if ($event =~ /\#0\#CF\((\d+)\,0\)\#/) {
		#           #0#CF(Clock_freq,0)# 
		$TIME_CONVERSION = 1_000_000_000 / $1;
	    } else {
		print STDERR "Error in second event";
		exit -1;
	    }  

	    # The following event should be #timestamp#TR(0,P)#.  The
	    # task 0 (main task) is the running task at the very
	    # beginning so this event is redundant
	    ($event, @events) = @events;
	    if ($event =~ /\#\d+\#TR\(0\,(\d+)\)\#/) {
		$main_initial_prio = $1;
	    } else {
		print STDERR "Error in third event";
		exit -1;
	    }
	    last; # leave the loop
	}
    } else {
	print STDERR "Error looking for first event";
	exit (-1);
    }
}
# initialize data
for ($i = 0; $i<$NTASKS + $NMUTEXES; $i++) {
    $values[$i] = "";
    $last_value[$i] = 0;
    $id_used[$i] = 0;
    $ready[$i] = 0;
}
$ready[0] = 1; # The main task starts ready
$ready[$NTASKS-1] = 1; # The idle task starts ready

# The idle tasks could be silent but it always exists
$id_used[$NTASKS-1] = 1;
$last_value[$NTASKS-1] = $BLOCKED + $NTASKS-1;
$values[$NTASKS-1] .= "0  " . ($last_value[$NTASKS-1]) . "\n";
# The main task starts running
$running_id = 0;
$id_used[0] = 1;
$last_value[0] = $RUNNING + $main_initial_prio*$PRIO_GAP;
$values[0] .= "0  " . ($last_value[0]) . "\n";

# Analice each scheduling event
foreach $event (@events) {

    if ($event =~ /\#(\d+)\#/) {
	# Take the timestamp part
	$ts = substr ($1, -12, -3);
	$event =~ s/\#(\d+)\#/\#/;
	if ($is_the_first) {
	    $is_the_first = 0;
	    $firts_timestamp = $ts;	
	}    
	$ts -= $firts_timestamp;
	$ts *= $TIME_CONVERSION;
    
	if ($event =~ /\#NT\((\d+)\,(\d+)\)\#/) {
	    # New Task #NT(id.....,prio.)#
	    $id = $1; $prio = $2;
	    if ($id_used[$id]) {
		$values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    }
	    $id_used[$id] = 1;
	    $last_value[$id] = $BLOCKED + $id;
	    $values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    $ready[$id] = 0;
	    
	} elsif ($event =~ /\#TR\((\d+)\,(\d+)\)\#/) {
	    # Task Ready    #TR(id.....,prio.)#
	    $id = $1; $prio = $2;
	    if ($id_used[$id]) {
		$values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    }
	    $id_used[$id] = 1; 
	    $last_value[$id] = $READY + $prio*$PRIO_GAP + $id;
	    $values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    $ready[$id] = 1;
	    
	} elsif ($event =~ /\#TP\((\d+)\,(\d+)\)\#/) {
	    # Task Raises Priority  #TP(id,new_prio.)#
	    $id = $1; $prio = $2;
	    if ($id_used[$id]) {
		$values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    }
	    $id_used[$id] = 1;
	    if ($running_id == $id) {
		$last_value[$id] = $RUNNING + $prio*$PRIO_GAP + $id;
	    } elsif ($ready[$id]) {
		$last_value[$id] = $READY + $prio*$PRIO_GAP + $id;
	    }
	    $values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    
	} elsif ($event =~ /\#Tp\((\d+)\,(\d+)\)\#/) {
	    # Task Reduces Priority  #Tp(id,new_prio.)#
	    $id = $1; $prio = $2;
	    if ($id_used[$id]) {
		$values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    }
	    $id_used[$id] = 1;
	    if ($running_id == $id) {
		$last_value[$id] = $RUNNING + $prio*$PRIO_GAP + $id;
	    } elsif ($ready[$id]) {
		$last_value[$id] = $READY + $prio*$PRIO_GAP + $id;
	    }
	    $values[$id] .= "$ts  " . ($last_value[$id]) . "\n";

	} elsif ($event =~ /\#TB\((\d+)\,(\d+)\)\#/) {
	    # Task Blocks   #TB(id.....,prio.)#
	    $id = $1; $prio = $2;
	    if ($id_used[$id]) {
		$values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    } 
	    $id_used[$id] = 1;
	    $last_value[$id] = $BLOCKED + $id;
	    $values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    $ready[$id] = 0;

	} elsif ($event =~ /\#TT\((\d+)\,(\d+)\)\#/) {
	    # Task Terminates #TT(id.....,prio.)#
	    $id = $1; $prio = $2;
	    if ($id_used[$id]) {
		$values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    } 
	    $id_used[$id] = 1;
	    $last_value[$id] = $TERMINATED + $id;
	    $values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    $ready[$id] = 0;

	} elsif ($event =~ /\#CS\((\d+)\,(\d+)\)\->\((\d+)\,(\d+)\)\#/){
	    # ContextSwitch #CS(id_old.,prio_old)->(id_new.,pri_new)#
	    $id_old = $1; $prio_old = $2; $id_new = $3; $prio_new = $4;
	    if ($id_used[$id_old]) {
		$values[$id_old] .= "$ts  " . ($last_value[$id_old]) . "\n";
	    } 
	    if ($id_used[$id_new]) {
		$values[$id_new] .= "$ts  " . ($last_value[$id_new]) . "\n";
	    }
	    $id_used[$id_old] = 1; $id_used[$id_new] = 1; 
	    if ($ready[$id_old]) {
		$last_value[$id_old] = 
		    $READY + $prio_old*$PRIO_GAP + $id_old;
		$values[$id_old] .= "$ts  " . ($last_value[$id_old]) . "\n";
	    }
	    $last_value[$id_new] = $RUNNING + $prio_new*$PRIO_GAP + $id_new;
	    $values[$id_new] .= "$ts  " . ($last_value[$id_new]) . "\n";
	    $running_id = $id_new;
    
	} elsif ($event =~ /\#NM\((\d+)\,(\d+)\)\#/) {
	    # New Mutex      #NM(id.....,prio.)#
	    $id = $1 + $NTASKS; $prio = $2;
	    if ($id_used[$id]) {
		$values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    }
	    $id_used[$id] = 1;
	    $last_value[$id] = $MUTEXES + (2*$PRIO_GAP*($id-$NTASKS));
	    $values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    
	} elsif ($event =~ /\#LM\((\d+)\,(\d+)\)\#/) {
	    # Lock Mutex     #LM(id.....,prio.)#
	    $id = $1 + $NTASKS; $prio = $2;
	    if ($id_used[$id]) {
		$values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    }
	    $id_used[$id] = 1; 
	    $last_value[$id] = 
		$MUTEXES + (2*$PRIO_GAP*($id-$NTASKS)) + $PRIO_GAP;
	    $values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    
	} elsif ($event =~ /\#IM\((\d+)\,(\d+)\)\#/) {
	    # Inherit Mutex  #IM(id.....,prio.)#
	    $id = $1 + $NTASKS; $prio = $2;
	    if ($id_used[$id]) {
		$values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    }
	    $id_used[$id] = 1; 
	    $last_value[$id] = 
		$MUTEXES + (2*$PRIO_GAP*($id-$NTASKS)) + $PRIO_GAP;
	    $values[$id] .= "$ts  " . ($MUTEXES + 
				       (2*$PRIO_GAP*($id-$NTASKS)) +
				       $PRIO_GAP/2) . "\n";
	    $values[$id] .= "$ts  " . ($last_value[$id]) . "\n";

	} elsif ($event =~ /\#UM\((\d+)\,(\d+)\)\#/) {
	    # Unlock Mutex   #UM(id.....,prio.)#
	    $id = $1 + $NTASKS; $prio = $2;
	    if ($id_used[$id]) {
		$values[$id] .= "$ts  " . ($last_value[$id]) . "\n";
	    } 
	    $id_used[$id] = 1;
	    $last_value[$id] = $MUTEXES + (2*$PRIO_GAP*($id-$NTASKS));
	    $values[$id] .= "$ts  " . ($last_value[$id]) . "\n";

	} else {
	    print STDERR "Error 2 in event:$event";
	    last;
	}
    } else {
	print STDERR "Error 1 in event:$event" unless $event =~ "Connected";
	last;
    }
}

#
# write results to files and a commands file for gnuplot
#
open COMM_FILE, ">comm.ti" or die "Can't open com.ti";
print COMM_FILE "plot "; # start plot ... command

# names of the outcome files
$out_file[0] = "T.main";
for ($i=1; $i<$NTASKS-1; $i++) {
    $out_file[$i] = "T.$i";
}
$out_file[$NTASKS-1] = "T.idle";
for ($i=0; $i<$NMUTEXES; $i++) {
    $out_file[$i+$NTASKS] = "M.$i";
}


# write one file per task or mutex
for ($i=0; $i<$NTASKS+$NMUTEXES; $i++) {
    if ($id_used[$i]) {
	print COMM_FILE "," unless $i == 0;

	open OUT_FILE, ">$out_file[$i]" or die "Can't open $out_file[$i]";
	print COMM_FILE "\"$out_file[$i]\" with lines";

        # repeat the last value in the final time
	$values[$i] =~ m/(\-*\d+)\n$/s;
	$values[$i] = $values[$i] . "$ts  $1\n";
	
	print OUT_FILE $values[$i] . "\n";
	close OUT_FILE;
    }
}
print COMM_FILE "\n"; # new line after the plot ... command

# write events at the bottom
close EVENTS_FILE;
open EVENTS_FILE, "<$file_name" or die "Can't open $file_name";
@events = <EVENTS_FILE>;
$i = 0;
$skip_2 = 2; # to skip the two first lines
foreach $event (@events) {
    if ($event =~ /\#(\d+)\#([A-Za-z]{2}.+)\#/) {
	if ($skip_2) {
	    $skip_2--;
	} else {
	    $ts = substr ($1, -12, -3);
	    print COMM_FILE "set label \"$2\" at first ",  
                  ($ts - $firts_timestamp) * $TIME_CONVERSION,",", 
	          $TERMINATED - 500 - $i,"\n";
	    $i = ($i + 500) % 3500;
	}
    }
}
	
#
# write tics in the 'y-axe'
#
print COMM_FILE "set ytics \(\"TERMINATED\" ", $TERMINATED,",\"BLOCKED\" ", $BLOCKED,",\"READY\" ",$READY;
# ready priorities
for($i=2; $i<$NPRIOS; $i+=2) {
   print COMM_FILE ",\"", $i, "\" ", $READY + $i *$PRIO_GAP;
}
# running
print COMM_FILE ",\"RUNNING\" ", $RUNNING; 
for($i=2; $i<$NPRIOS; $i+=2) {
   print COMM_FILE ",\"", $i, "\" ", $RUNNING + $i *$PRIO_GAP;
}
# mutexes 
print COMM_FILE ",\"MUTEXES\" ", $MUTEXES;
for ($i=$NTASKS+1; $i<$NTASKS+$NMUTEXES; $i++) {
    if ($id_used[$i]) {
	print COMM_FILE ",\"", $i-$NTASKS, "\" ", $MUTEXES+ (2*$PRIO_GAP*($i-$NTASKS));
    }
} 
print COMM_FILE "\)\n"; 

# show full range
print COMM_FILE "set yrange \[",-6000,":\]\n"; 
  
# show grid
print COMM_FILE "set grid\n";   

print COMM_FILE "replot";
close COMM_FILE;

system "gnuplot comm.ti \"-\"" or die "Executing gnuplot comm.ti&";





