#!/usr/bin/perl -w

# use:
#  $ run_tests.pl [-ddir_tests] [-c]

# -d
#    Directory in which can be found the tests. If no
#    -d option is specified $MPATH/tests will be used

# -c
#    Compile only (don't execute the tests)


require "root_instalation_path/utils/globals.pl";
use vars qw($GNAT_VERSION $ARCH $GNAT_PATH);


# exec_command (cmd)
sub exec_command;

# exec_command_no_echo (cmd)
sub exec_command_no_echo;

# test_for_marte_arch (file, arch)
sub test_for_marte_arch;

# print_log (msg)
# append msg to the log file
sub print_log;

my $VERVOSE=1;

# ---------------------------------------------------------------------
# Main
#

# ------------------- #
# Argumets processing #
# ------------------- #
use Getopt::Long;
Getopt::Long::Configure(qw (pass_through no_auto_abbrev no_ignore_case) );
my $base_tests_dir = "";
my $compile_only=0;
GetOptions("d=s" => \$base_tests_dir,
           "c"   => \$compile_only);

if ($base_tests_dir eq "") {
    $base_tests_dir="$MPATH/tests";
} elsif (!($base_tests_dir =~ /^\//)) {
    # relative path
    $base_tests_dir="$MPATH/tests/$base_tests_dir";
}

(-e $base_tests_dir) or die "Couldn't find dir: $base_tests_dir";

# ----------------------------- #
#  Get MaRTE arch/GNAT_version  #
# ----------------------------- #
$M_ARCH_GNAT_VER=$ARCH . "/" . $GNAT_VERSION;

# ------------ #
#   Log file   #
# ------------ #
$log_file="$base_tests_dir/00tests.log";
exec_command_no_echo "rm -f $log_file";
print_log "Running tests on $base_tests_dir\n";

# --------------- #
#    Run tests    #
# --------------- #
my $global_tests_count=0;
my $global_fail_count=0;

print "  -> Regression tests in $base_tests_dir\n";

# Run tests in every directory
opendir DIR, "$base_tests_dir" or die "directory $base_tests_dir not found";
@all_dirs = grep !/^\./, readdir DIR; # remove files starting with "."
push (@all_dirs, "."); # add "." directory to the list
foreach $dir (@all_dirs) {
  $tests_dir = "$base_tests_dir/$dir";
  # only directories:
  next unless (-d "$tests_dir" && 
	       $dir ne "comp_errors" && $dir ne "exec_errors");

  my $tests_count=0;
  my $fail_count=0;

  # remove files from former tests
  exec_command_no_echo "cd $tests_dir && " .
      "make clean >> $log_file 2>&1";

  #  Set correct reports-init.adb (only for Ada tests)
  if (-e "$tests_dir/reports") {
      if ($ARCH eq "x86") {
	  exec_command_no_echo "cd $tests_dir/reports && " .
	      "cp reports-init.adb.x86 reports-init.adb";
      } else {
	  exec_command_no_echo "cd $tests_dir/reports && " .
	      "cp reports-init.adb.linux reports-init.adb";
      }
  }

  # Compile and execute every file in directory
  opendir DIR, "$tests_dir" or die "directory $tests_dir not found";
  @all_files = grep !/^\./, readdir DIR;
  foreach $file (@all_files) {
    next unless ($file =~ /^test_/);

    # Check if it is a test for this architecture
    $test_for_this_arch = test_for_marte_arch("$tests_dir/$file",
                          $M_ARCH_GNAT_VER);
    $tests_count=$tests_count+1 unless !$test_for_this_arch;
    if ($test_for_this_arch == -1) {
    print "  -> " . " $file: Not found \"--  Test [not] for: ...\"\n";
    $fail_count=$fail_count+1;
    next;
    }
    next unless $test_for_this_arch;

    # get file name without ".adb" or ".c"
    $file_no_adb=$file;
    $file_no_adb=~s/\.adb$//;
    $file_no_adb=~s/\.c$//;
    $file_no_adb=~s/\.cpp$//;

    my $msg_test = "  $file: compile: ";

    # ------------------
    # compile file
    print_log "\n--  $file: ------------------------------------------\n";

    if (system("cd $tests_dir && " .
           "export PATH=$GNAT_PATH/bin:\$PATH:$MPATH/utils && " .
           "make $file_no_adb.exe >> $log_file 2>&1") != 0) {
    print "  ->" . $msg_test . "<-- FAIL (compilation)\n";
    $fail_count=$fail_count+1;
    next;
    }

    # look for executable
    if (-e "$tests_dir/$file_no_adb.exe") {
        $executable="$file_no_adb.exe";
    } else {
        print "  ->" . $msg_test . "<-- FAIL (not executable found)\n";
        $fail_count=$fail_count+1;
        next;
    }
    print $msg_test . "OK\n" unless $VERVOSE==0;  # compilation OK

    if ($compile_only) {
        # nothing more to be done for this file
        next;
    }

    # ------------------
    # execute file
    $msg_test = "  $file: execute: ";

    if ($ARCH eq "x86") {
	# get user's name
	open (USR, "whoami |");
	my $user=<USR>;
	$user =~ s/\n//;
	close USR;

	# launch QEMU
	print_log "Launching QEMU for $EXPORTED_DIR/disk.img\n";
        open (SYS_OUT, "mcopy -o $tests_dir/$executable c:mprogram && " .
	      "cd $EXPORTED_DIR && " .
	      "timeout 20 " .
	      "/usr/local/bin/qemu -hda disk.img -nographic -serial stdio 2>&1 |");
    } else {
        open (SYS_OUT, "cd $tests_dir && ./$executable 2>&1 |");
    }
    @mprogram_out = <SYS_OUT>;

    print_log "--  $file execution output: --\n" .
	"\n@mprogram_out\n";

    # ------------------
    # parse output
    $test_ok=0;
    $assert_fail=0;
    foreach $line (@mprogram_out) {
    #print "$line";
	if ($line =~ /ASSERT FAILED/) {
	    $assert_fail=1;
	    if (!($file =~ /test_assert_fail/)) {
		print "  ->" . $msg_test . "<--  FAIL (assert)\n";
		$fail_count=$fail_count+1;
	    } else {
		# expected faliure
		print $msg_test . "EXPECTED ASSERT FAIL\n" unless $VERVOSE==0;
	    }
	    next;
	} elsif ($line =~ /Test OK/) {
	    print $msg_test . "OK\n" unless $VERVOSE==0;
	    $test_ok=1;
	}
    }
    close SYS_OUT;

    # Test OK missed??  (only if not assert fail)
    if (!$assert_fail && !$test_ok) {
    if (!($file =~ /test_not_test_ok/)) {
        print "  ->" . $msg_test . "<--  FAIL (\"Test OK\" not found)\n";
        $fail_count=$fail_count+1;
    } else {
        # expected faliure
        print $msg_test . "EXPECTED \"Test OK\" FAIL\n" unless $VERVOSE==0;
        $test_ok=1;
    }
    }


  } # foreach file
  printf "  -> TEST RESULT (%15s/): ", $dir; 
  print  "" . ($fail_count) . "/" . ($tests_count) . "  (fail/run)\n\n";
  $global_tests_count += $tests_count;
  $global_fail_count  += $fail_count;
} # foreach dir
print "  -> GLOBAL TEST RESULT: ". ($global_fail_count) . "/" . ($global_tests_count) . "  (fail/run)\n";

exit 0;

# -----------------------------------------------------------------------------
# Subrutines
# -----------------------------------------------------------------------------

# test_for_marte_arch (file, arch)
sub test_for_marte_arch
{
  my $file = shift(@_);
  my $arch = shift(@_);
  #print "file:$file arch:$arch\n";
  open F, "$file" or die "Couldn't open $file\n";

  # Get the first line
  my $line = <F>;
  close F;

  if ($line =~ /Test for all architectures/) {
      return 1;
  }

  if ($line =~ /Test for:/) {
      return ($line =~ /$arch/);
  } elsif ($line =~ /Test not for:/) {
      return !($line =~ /$arch/);
  }
  return -1;
}

# print_lob (msg)
sub print_log
{
    my $msg = shift(@_);
    open LOG_FILE, ">>$log_file";
    print LOG_FILE "$msg";
    close LOG_FILE;
}

# exec_command (cmd)
sub exec_command
{
        my $cmd = shift(@_);
        printf "$cmd\n";
        system ($cmd) == 0 or die "Can't execute $cmd\n";
}

# exec_command_no_echo (cmd)
sub exec_command_no_echo
{
        my $cmd = shift(@_);
        system ($cmd) == 0 or die "Can't execute $cmd\n";
}
    

